//! This module contains the code necessary to run and supervise actors.

use std::{future::Future, pin::Pin};

use log::{debug, error, warn};
use tokio::{
    select,
    sync::{mpsc, oneshot},
    task::{AbortHandle, JoinSet},
};

use crate::{ActorExit, ActorFault, ActorId, ControlMsg, Runtime};

pub(super) type FaultResult = std::result::Result<ActorId, ActorFault>;
pub(super) type BoxedFuture = Pin<Box<dyn Send + Future<Output = FaultResult>>>;

pub struct SpawnReq {
    task: BoxedFuture,
    sender: oneshot::Sender<AbortHandle>,
}

impl SpawnReq {
    pub(super) fn new<Fut>(task: Fut) -> (Self, oneshot::Receiver<AbortHandle>)
    where
        Fut: 'static + Send + Future<Output = FaultResult>,
    {
        let (sender, receiver) = oneshot::channel();
        let req = Self {
            task: Box::pin(task),
            sender,
        };
        (req, receiver)
    }
}

/// The kernel is responsible for spawning and supervising actors.
pub(super) async fn kernel(runtime: &Runtime, mut tasks: mpsc::Receiver<SpawnReq>) {
    let mut running = JoinSet::<FaultResult>::new();

    loop {
        select! {
            option = tasks.recv() => {
                match option {
                    Some(SpawnReq { task, sender }) => {
                        let handle = running.spawn(task);
                        if sender.send(handle).is_err() {
                            error!("SpawnReq sender dropped the channel. Aborting actor.");
                        }
                    }
                    None => panic!("The Runtime was dropped!")
                };
            }
            Some(result) = running.join_next() => {
                match result {
                    Ok(actor_result) => {
                        handle_actor_exit(actor_result, runtime).await;
                    }
                    Err(join_error) => {
                        match join_error.try_into_panic() {
                            Ok(panic_obj) => {
                                if let Some(message) = panic_obj.downcast_ref::<String>() {
                                    error!("An actor panic was uncaught: {message}");
                                } else {
                                    error!("An actor panic was uncaught.");
                                }
                            }
                            Err(join_error) => {
                                debug!("Actor was aborted: {join_error}");
                            }
                        };
                    }
                }
            }
        }
    }
}

/// Handles the termination of an actor by notifying the actor its owned by and all the actors it
/// owns.
async fn handle_actor_exit(actor_result: FaultResult, runtime: &Runtime) {
    let (actor_id, actor_exit) = match actor_result {
        Ok(actor_id) => {
            debug!("Actor {actor_id} exited normally.");
            (actor_id, ActorExit::Ok)
        }
        Err(err) => {
            error!("{err}");
            (err.actor_id(), ActorExit::from(err))
        }
    };

    let mut handle = {
        let mut handles = runtime.handles.write().await;
        if let Some(handle) = handles.remove(&actor_id) {
            handle
        } else {
            warn!("Exiting actor {actor_id} was has been taken from the runtime before notification could be sent.");
            return;
        }
    };

    if let Some(owner) = handle.take_owner() {
        let handle_name = handle.name();
        let msg = ControlMsg::OwnedExited {
            name: handle_name.clone(),
            exit: actor_exit.clone(),
        };
        let result = runtime.send_control(owner.clone(), msg).await;
        if let Err(err) = result {
            error!("Failed to notify owner {owner} that {handle_name} exited: {err}");
        }
    }

    let handle_name = handle.name().clone();
    for owned in handle.owns_mut().drain(0..) {
        let msg = ControlMsg::OwnerExited(actor_exit.clone());
        let result = runtime.send_control(owned.clone(), msg).await;
        if let Err(err) = result {
            error!("Failed to notify owned actor {owned} that {handle_name} exited: {err}");
        }
    }
}
