//! This module contains types used to model the actor system implemented by the runtime.

use std::{any::Any, fmt::Display, future::Future, pin::Pin, sync::Arc};

use btlib::{bterr, BlockPath, Result};
use btserde::field_helpers::smart_ptr;
use serde::{de::DeserializeOwned, Deserialize, Serialize};
use tokio::{
    sync::{mpsc, oneshot},
    task::AbortHandle,
};
use uuid::Uuid;

use crate::WireEnvelope;

pub use once_cell::sync::Lazy;

/// The return type of actor closures.
pub type ActorResult = std::result::Result<ActorId, ActorError>;

pub trait Named {
    fn name(&self) -> Arc<String>;
}

/// Represents the result of an actor state transition, which can be one of the following:
/// * `Success`: The transition succeeded and the new state is contained in the result.
/// * `Abort`: The transition failed and the previous state along with an error describing
///   the failure is in the result.
/// * `Fatal`: The transition failed and the actor cannot recover from this failure. An error
///   describing the failure is contained in the result.
pub enum TransResult<From, To> {
    /// Represents a successful transition.
    Ok(To),
    /// Represents an aborted transition.
    Abort { from: From, err: btlib::Error },
    /// Represents a failed transition which kills the actor which attempted it.
    Fatal { err: btlib::Error },
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
/// Specifies a kind of transition, either a `Send` or a `Receive`.
pub enum TransKind {
    /// A transition which doesn't receive any message, but sends one or more.
    Send,
    /// A transition which receives a message.
    Receive,
}

impl TransKind {
    const fn verb(&self) -> &'static str {
        match self {
            Self::Send => "sending",
            Self::Receive => "receiving",
        }
    }
}

impl Copy for TransKind {}

#[derive(Debug, Clone, Serialize, Deserialize)]
/// A struct which conveys information about where an actor error occurred to the kernel.
pub struct ActorErrorCommon {
    /// The ID of the actor which experienced the error.
    pub actor_id: ActorId,
    /// The name of the actor implementation which erred.
    #[serde(with = "smart_ptr")]
    pub actor_impl: Arc<String>,
    /// The name of the state the actor was in.
    #[serde(with = "smart_ptr")]
    pub state: Arc<String>,
    /// The name of the message the actor was handling, or the name of the first message it was
    /// trying to send.
    #[serde(with = "smart_ptr")]
    pub message: Arc<String>,
    /// The kind of transition which was being attempted.
    pub kind: TransKind,
}

impl Display for ActorErrorCommon {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "Actor {}, with implementation '{}', panicked in state {} while {} a message of type {}",
            self.actor_id,
            self.actor_impl,
            self.state,
            self.kind.verb(),
            self.message
        )
    }
}

pub struct ActorError {
    common: ActorErrorCommon,
    err: btlib::Error,
}

impl ActorError {
    pub fn new(err: btlib::Error, common: ActorErrorCommon) -> Self {
        Self { common, err }
    }

    pub fn actor_id(&self) -> ActorId {
        self.common.actor_id
    }

    pub fn actor_impl(&self) -> &Arc<String> {
        &self.common.actor_impl
    }

    pub fn state(&self) -> &Arc<String> {
        &self.common.state
    }

    pub fn message(&self) -> &Arc<String> {
        &self.common.message
    }

    pub fn kind(&self) -> TransKind {
        self.common.kind
    }

    pub fn err(&self) -> &btlib::Error {
        &self.err
    }
}

impl Display for ActorError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}: {}", self.common, self.err)
    }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct ActorErrorStr {
    common: ActorErrorCommon,
    #[serde(with = "smart_ptr")]
    err: Arc<String>,
}

impl ActorErrorStr {
    pub fn actor_id(&self) -> ActorId {
        self.common.actor_id
    }

    pub fn actor_impl(&self) -> &Arc<String> {
        &self.common.actor_impl
    }

    pub fn state(&self) -> &Arc<String> {
        &self.common.state
    }

    pub fn message(&self) -> &Arc<String> {
        &self.common.message
    }

    pub fn kind(&self) -> TransKind {
        self.common.kind
    }

    pub fn err(&self) -> &Arc<String> {
        &self.err
    }
}

impl Display for ActorErrorStr {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}: {}", self.common, self.err)
    }
}

impl From<ActorError> for ActorErrorStr {
    fn from(value: ActorError) -> Self {
        let err = Arc::new(value.err.to_string());
        Self {
            common: value.common,
            err,
        }
    }
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub enum PanicKind {
    String(#[serde(with = "smart_ptr")] Arc<String>),
    Unknown,
}

impl PanicKind {
    fn new(err: Box<dyn Any>) -> Self {
        if let Ok(err) = err.downcast::<String>() {
            let string = *err;
            Self::String(Arc::new(string))
        } else {
            Self::Unknown
        }
    }
}

impl Display for PanicKind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::String(string) => f.write_str(string.as_str()),
            Self::Unknown => f.write_str("Unknown error, panic object was not a String."),
        }
    }
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ActorPanic {
    actor_id: ActorId,
    err: PanicKind,
}

impl ActorPanic {
    pub(super) fn new(actor_id: ActorId, err: Box<dyn Any>) -> Self {
        Self {
            actor_id,
            err: PanicKind::new(err),
        }
    }

    pub fn actor_id(&self) -> ActorId {
        self.actor_id
    }

    pub fn err(&self) -> Option<&str> {
        if let PanicKind::String(string) = &self.err {
            Some(string.as_str())
        } else {
            None
        }
    }
}

impl Display for ActorPanic {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let actor_id = &self.actor_id;
        let err = &self.err;
        write!(f, "Actor {actor_id} panicked: {err}")
    }
}

/// Represents the ways an actor can exit abnormally, either by returning an error or panicking.
pub(super) enum ActorFault {
    Error(ActorError),
    Panic(ActorPanic),
}

impl ActorFault {
    pub(crate) fn actor_id(&self) -> ActorId {
        match self {
            Self::Error(actor_error) => actor_error.actor_id(),
            Self::Panic(ActorPanic { actor_id, .. }) => *actor_id,
        }
    }
}

impl Display for ActorFault {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Error(err) => err.fmt(f),
            Self::Panic(panic) => panic.fmt(f),
        }
    }
}

#[derive(Clone, Serialize, Deserialize)]
/// Represents the ways an actor can exit.
pub enum ActorExit {
    /// The actor exited normally.
    Ok,
    /// The actor exited with an error.
    Error(ActorErrorStr),
    /// The actor panicked.
    Panic(ActorPanic),
}

impl From<ActorFault> for ActorExit {
    fn from(value: ActorFault) -> Self {
        match value {
            ActorFault::Error(error) => Self::Error(error.into()),
            ActorFault::Panic(panic) => Self::Panic(panic),
        }
    }
}

/// Represents the terminal state of an actor, where it stops processing messages and halts.
pub struct End;

impl End {
    /// Returns the identifier for this type which is expected in protocol definitions.
    pub fn ident() -> &'static str {
        stringify!(End)
    }
}

/// A unique identifier for a particular service.
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ServiceId(#[serde(with = "smart_ptr")] Arc<String>);

impl ServiceId {
    pub fn new(value: Arc<String>) -> Self {
        Self(value)
    }
}

impl AsRef<str> for ServiceId {
    fn as_ref(&self) -> &str {
        self.0.as_str()
    }
}

impl<T: Into<String>> From<T> for ServiceId {
    fn from(value: T) -> Self {
        Self(Arc::new(value.into()))
    }
}

impl Display for ServiceId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.as_ref())
    }
}

/// A unique identifier for a service.
///
/// A service is a collection of actors in the same directory which provide some functionality.
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ServiceName {
    /// The path to the directory containing the service.
    #[serde(with = "smart_ptr")]
    path: Arc<BlockPath>,
    /// The id of the service.
    service_id: ServiceId,
}

impl ServiceName {
    pub fn new(path: Arc<BlockPath>, service_id: ServiceId) -> Self {
        Self { path, service_id }
    }
}

/// Indicates the set of service providers a message is addressed to.
///
/// The message could be delivered to any of the service providers in this set, at random.
pub struct ServiceAddr {
    /// The [ServiceName] to address the message to.
    name: ServiceName,
    #[allow(dead_code)]
    /// Indicates if the message should be routed towards the root of the tree or away from it.
    rootward: bool,
}

impl ServiceAddr {
    pub fn new(name: ServiceName, rootward: bool) -> Self {
        Self { name, rootward }
    }

    pub fn path(&self) -> &Arc<BlockPath> {
        &self.name.path
    }

    pub fn service_id(&self) -> &ServiceId {
        &self.name.service_id
    }
}

/// An identifier for an actor which is unique in a given runtime.
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ActorId(Uuid);

impl ActorId {
    pub fn new() -> Self {
        Self(Uuid::new_v4())
    }

    /// Returns an actor ID that can't have messages delivered to it.
    pub fn undeliverable() -> Self {
        Self(Uuid::from_bytes([0; 16]))
    }
}

impl Default for ActorId {
    fn default() -> Self {
        Self::new()
    }
}

impl Copy for ActorId {}

impl Display for ActorId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

/// A unique identifier for a specific actor activation.
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ActorName {
    /// The path to the directory containing this actor.
    #[serde(with = "smart_ptr")]
    path: Arc<BlockPath>,
    /// A unique identifier for an actor activation. Even as an actor transitions to different types
    /// as it handles messages, this value does not change. Thus this value can be used to trace an
    /// actor through a series of state transitions.
    act_id: ActorId,
}

impl ActorName {
    pub fn new(path: Arc<BlockPath>, act_id: ActorId) -> Self {
        Self { path, act_id }
    }

    pub fn path(&self) -> &Arc<BlockPath> {
        &self.path
    }

    pub fn act_id(&self) -> ActorId {
        self.act_id
    }
}

impl Display for ActorName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}@{}", self.act_id, self.path)
    }
}

/// Trait for messages which expect exactly one reply.
pub trait CallMsg: Serialize + DeserializeOwned + Send + Sync {
    /// The reply type expected for this message.
    type Reply: Serialize + DeserializeOwned + Send + Sync;
}

/// Trait for messages which expect exactly zero replies.
pub trait SendMsg: CallMsg {}

pub trait MsgEnum: CallMsg + Named {}

impl<T: CallMsg + Named> MsgEnum for T {}

/// A type used to express when a reply is not expected for a message type.
#[derive(Serialize, Deserialize)]
pub enum NoReply {}

/// Messages sent by the runtime to actors.
pub enum ControlMsg {
    OwnerExited(ActorExit),
    OwnedExited { name: ActorName, exit: ActorExit },
}

impl Named for ControlMsg {
    fn name(&self) -> Arc<String> {
        static OWNER_EXITED_NAME: Lazy<Arc<String>> = Lazy::new(|| Arc::new("OwnerExited".into()));
        static OWNED_EXITED_NAME: Lazy<Arc<String>> = Lazy::new(|| Arc::new("OwnedExited".into()));
        match self {
            Self::OwnerExited(_) => OWNER_EXITED_NAME.clone(),
            Self::OwnedExited { .. } => OWNED_EXITED_NAME.clone(),
        }
    }
}

/// Wrapper around a message type `T` which indicates who the message is from and, if the message
/// was dispatched with `call`, provides a channel to reply to it.
pub enum Envelope<T: MsgEnum> {
    Call {
        msg: T,
        reply: Option<oneshot::Sender<T::Reply>>,
    },
    Send {
        msg: T,
        from: ActorName,
    },
    Control(ControlMsg),
}

impl<T: MsgEnum> Envelope<T> {
    /// Creates a new envelope containing the given message which does not expect a reply.
    pub(super) fn new_send(from: ActorName, msg: T) -> Self {
        Self::Send { from, msg }
    }

    /// Creates a new envelope containing the given message which expects exactly one reply.
    pub(super) fn new_call(msg: T) -> (Self, oneshot::Receiver<T::Reply>) {
        let (tx, rx) = oneshot::channel::<T::Reply>();
        let envelope = Self::Call {
            reply: Some(tx),
            msg,
        };
        (envelope, rx)
    }

    /// Returns the name of the actor which sent this message.
    pub fn from(&self) -> Option<&ActorName> {
        match self {
            Self::Send { from, .. } => Some(from),
            _ => None,
        }
    }

    /// Returns a reference to the message in this envelope.
    pub fn msg(&self) -> Option<&T> {
        match self {
            Self::Call { msg, .. } | Self::Send { msg, .. } => Some(msg),
            _ => None,
        }
    }

    pub const fn name(&self) -> &'static str {
        match self {
            Self::Call { .. } => "Call",
            Self::Send { .. } => "Send",
            Self::Control { .. } => "Control",
        }
    }

    pub fn msg_name(&self) -> Arc<String> {
        match self {
            Self::Call { msg, .. } | Self::Send { msg, .. } => msg.name(),
            Self::Control(msg) => msg.name(),
        }
    }

    /// Sends a reply to this message.
    ///
    /// If this message is not expecting a reply, or if this message has already been replied to,
    /// then an error is returned.
    pub fn reply(&mut self, reply: T::Reply) -> Result<()> {
        match self {
            Self::Call { reply: tx, .. } => {
                if let Some(tx) = tx.take() {
                    tx.send(reply).map_err(|_| bterr!("Failed to send reply."))
                } else {
                    Err(bterr!("Reply has already been sent."))
                }
            }
            _ => Err(bterr!("Can't reply to {} messages.", self.name())),
        }
    }

    /// Returns true if this message expects a reply and it has not already been replied to.
    pub fn needs_reply(&self) -> bool {
        matches!(self, Self::Call { .. })
    }
}

pub(super) type FutureResult = Pin<Box<dyn Send + Future<Output = Result<()>>>>;

pub struct ActorHandle {
    name: ActorName,
    handle: AbortHandle,
    sender: Box<dyn Send + Sync + Any>,
    deliverer: Box<dyn Send + Sync + Fn(WireEnvelope<'_>) -> FutureResult>,
    ctrl_deliverer: Box<dyn Send + Sync + Fn(ControlMsg) -> FutureResult>,
    owner: Option<ActorName>,
    owns: Vec<ActorName>,
}

impl ActorHandle {
    pub(super) fn new<T, F, G>(
        name: ActorName,
        handle: AbortHandle,
        sender: mpsc::Sender<Envelope<T>>,
        deliverer: F,
        ctrl_deliverer: G,
        owner: Option<ActorName>,
    ) -> Self
    where
        T: 'static + MsgEnum,
        F: 'static + Send + Sync + Fn(WireEnvelope<'_>) -> FutureResult,
        G: 'static + Send + Sync + Fn(ControlMsg) -> FutureResult,
    {
        Self {
            name,
            handle,
            sender: Box::new(sender),
            deliverer: Box::new(deliverer),
            ctrl_deliverer: Box::new(ctrl_deliverer),
            owner,
            owns: Vec::new(),
        }
    }

    fn sender<T: 'static + MsgEnum>(&self) -> Result<&mpsc::Sender<Envelope<T>>> {
        self.sender
            .downcast_ref::<mpsc::Sender<Envelope<T>>>()
            .ok_or_else(|| bterr!("Attempt to send message as the wrong type."))
    }

    pub fn name(&self) -> &ActorName {
        &self.name
    }

    /// Sends a message to the actor represented by this handle.
    pub async fn send<T: 'static + MsgEnum>(&self, from: ActorName, msg: T) -> Result<()> {
        let sender = self.sender()?;
        sender
            .send(Envelope::new_send(from, msg))
            .await
            .map_err(|_| bterr!("failed to enqueue message"))?;
        Ok(())
    }

    pub async fn call_through<T: 'static + MsgEnum>(&self, msg: T) -> Result<T::Reply> {
        let sender = self.sender()?;
        let (envelope, rx) = Envelope::new_call(msg);
        sender
            .send(envelope)
            .await
            .map_err(|_| bterr!("failed to enqueue call"))?;
        let reply = rx.await?;
        Ok(reply)
    }

    pub(crate) async fn send_control(&self, msg: ControlMsg) -> Result<()> {
        (self.ctrl_deliverer)(msg).await
    }

    pub fn abort(&self) {
        self.handle.abort();
    }

    pub(super) async fn deliver(&self, envelope: WireEnvelope<'_>) -> Result<()> {
        (self.deliverer)(envelope).await
    }

    pub(crate) fn take_owner(&mut self) -> Option<ActorName> {
        self.owner.take()
    }

    pub(crate) fn owns_mut(&mut self) -> &mut Vec<ActorName> {
        &mut self.owns
    }
}

impl Drop for ActorHandle {
    fn drop(&mut self) {
        self.abort();
    }
}

/// Expands to an implementation of an `actor_impl` method which returns reference to a statically
/// allocated name.
#[macro_export]
macro_rules! actor_name {
    ($name:literal) => {
        fn actor_impl() -> ::std::sync::Arc<String> {
            use ::once_cell::sync::Lazy;
            use ::std::sync::Arc;
            static NAME: Lazy<Arc<String>> = Lazy::new(|| Arc::new($name.into()));
            NAME.clone()
        }
    };
}
